/****************************************************************************
 *																			*
 *					RFM12 basierte, drahtlose RS232 Verbindung				*
 *																			*
 *								Version 1.1									*
 *																			*
 *								 by Benedikt								*
 *																			*
 *						Email:	benedikt83 t gmx.net						*
 *																			*
 ****************************************************************************
 *																			*
 *	Die Software darf frei kopiert und verndert werden, solange sie nicht	*
 *	ohne meine Erlaubnis fr kommerzielle Zwecke eingesetzt wird.			*
 *																			*
 ***************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include <portbits.h>
#include "global.h"
#include "uart.h"
#include "rf12.h"
#include "leds.h"

/*
DIP Schalter an PortC:
0-3: Baudrate: 1200,2400,9600,19200,38400,57600,115200 (hohe Baudraten nur mit entsprechendem Quarztakt mglich !)
4-5: HF Kanal
*/


#define F_CPU 10000000UL
#include <util/delay.h>

//#define DEBUG						// anstelle der empfangenen Daten fr jede Aktion den Status ausgeben

#define USE_DIPSWITCH				// DIP Schalter/Jumper an PortC 0-4
#define CHANNEL			1			// Sende/Empfangskanal (0-3) (nur gltig wenn kein DIP Schalter verwendet wird)
#define RF_BAUDRATE		20000		// Baudrate des RFM12 (nur gltig wenn kein DIP Schalter verwendet wird)
#define UART_BAUDRATE	19200		// Baudrate des UARTs (nur gltig wenn kein DIP Schalter verwendet wird)

#define MAX_BUF			128			// Paket Gre in Bytes (maximal 250)
#define TX_TIMEOUT		150			// Maximale Wartezeit auf Daten in ms (max 500)

#define ANSWER_TIMEOUT	10			// Maximale Wartezeit auf die Besttigung der Daten in ms (max 500)

#define RETRY			15			// Maximale Anzahl an Sendeversuchen

volatile unsigned char delaycnt;
unsigned char rxbuf[MAX_BUF+1];		// Puffer fr empfangene Daten
unsigned char txbuf[MAX_BUF+1];		// Puffer fr zu sendende Daten
unsigned char flags, tx_cnt, tx_id, tx_status, retrans_cnt;

#define WAITFORACK		1

void rx_packet(void);
void tx_packet(unsigned char retrans);

int main(void)
{
	PORTC=255;
	PORTD=3;
	DDRD=254;
	DDRC=0;

#ifdef USE_DIPSWITCH
	unsigned char c;
	unsigned short baud_uart, baud_rf;
	c=PINC&7;
	if (c==0)
	{	baud_rf=2000;
		baud_uart=UART_BAUD_SELECT(1200, F_CPU);
	}
	else if (c==1)
	{	baud_rf=4000;
		baud_uart=UART_BAUD_SELECT(2400, F_CPU);
	}
	else if (c==2)
	{	baud_rf=8000;
		baud_uart=UART_BAUD_SELECT(4800, F_CPU);
	}
	else if (c==3)
	{	baud_rf=15000;
		baud_uart=UART_BAUD_SELECT(9600, F_CPU);
	}
	else if (c==4)
	{	baud_rf=25000;
		baud_uart=UART_BAUD_SELECT(19200, F_CPU);
	}
	else if (c==5)
	{	baud_rf=30000;
		baud_uart=UART_BAUD_SELECT(38400, F_CPU);
	}
	else if (c==6)
	{	baud_rf=40000;
		baud_uart=UART_BAUD_SELECT(57600, F_CPU);
	}
	else if (c==7)
	{	baud_rf=50000;
		baud_uart=UART_BAUD_SELECT(115200, F_CPU);
	}
	c=(PINC&24)/8;
    uart_init(baud_uart);
	rf12_init();									// ein paar Register setzen (z.B. CLK auf 10MHz)
	rf12_config(baud_rf, c, 0, QUIET);				// Baudrate, Kanal (0-3), Leistung (0=max, 7=min), Umgebungsbedingungen (QUIET, NORMAL, NOISY)
#else
    uart_init(UART_BAUD_SELECT(UART_BAUDRATE, F_CPU));
	rf12_init();									// ein paar Register setzen (z.B. CLK auf 10MHz)
	rf12_config(RF_BAUDRATE, CHANNEL, 0, QUIET);	// Baudrate, Kanal (0-3), Leistung (0=max, 7=min), Umgebungsbedingungen (QUIET, NORMAL, NOISY)
#endif

	TCCR1A=0;
	TCCR1B=(1<<WGM12)|1;
	OCR1A=((F_CPU+2500)/500)-1;
	TIMSK=(1<<OCIE1A);

	sei();
	rf12_rxmode();

	for (;;)
	{	if (rf12_data())							// Daten im RF12 RX Puffer ?
		{	rx_packet();
		}
		if (!(flags&WAITFORACK))					// Puffer nur dann fllen, wenn die alten Daten aus dem Puffer wirklich angekommen sind
		{	if (uart_data())						// Daten im UART RX Puffer ?
			{	if (!tx_cnt)
					delaycnt=TX_TIMEOUT/2;			// Nach dem ersten Byte: timeout starten
				txbuf[tx_cnt++]=uart_getchar();
			}
			if ((tx_cnt>=MAX_BUF)||((tx_cnt)&&(delaycnt==0))) // Puffer voll, oder timeout seit erstem Byte im Puffer vorbei ?  -> senden
			{	tx_status=0;						// zu sendender Status
				tx_packet(0);						// erstmaliger Transfer
#ifdef DEBUG
				uart_putc('t');
#endif
			}
		}
		else if (delaycnt==0)						// Timeout: Daten nochmal senden
		{	LED_RETRANS=1;
			if (retrans_cnt)
			{	retrans_cnt--;
            	tx_packet(1);						// retransmit
#ifdef DEBUG
				uart_putc('r');
#endif
			}
			else									// Versuche abgelaufen
			{	LED_ERR=1;							// -> Fehler LED an
				tx_cnt=0;							// -> Daten verwerfen
				tx_id++;
				flags&=~WAITFORACK;					// Daten als OK markieren
#ifdef DEBUG
				uart_putc('e');
#endif

			}
		}
	}
}

void rx_packet(void)
{	static unsigned char rx_lastid=255;
	unsigned char rx_cnt,i, rx_id, status;
	rx_cnt=rf12_rxdata(rxbuf, &status, &rx_id);		// komplettes Paket empfangen
	if (rx_cnt<=MAX_BUF)							// Daten gltig (d.h. kein CRC Fehler) ?
	{	if (status&RECEIVED_OK)						// Empfangsbesttigung ?
		{	flags&=~WAITFORACK;						// -> "Warten auf Besttigung"-Flag lschen
			tx_cnt=0;								// -> Daten als gesendet markieren
			tx_id++;
			LED_RETRANS=0;
			LED_ERR=0;
#ifdef DEBUG
				uart_putc('a');
#endif
		}
		if (rx_cnt)									// Daten empfangen
		{	tx_status=RECEIVED_OK;					// zu sendender Status
#ifdef DEBUG
			uart_putc('p');
#endif
			tx_packet(0);							// Empfangsbesttigung senden
			retrans_cnt=RETRY;						// Retry Counter neu starten
			if (rx_id!=rx_lastid)					// Handelt es sich um neue Daten ?
			{	
#ifndef DEBUG
				for (i=0; i<rx_cnt; i++)			// Daten weiterleiten
					uart_putc(rxbuf[i]);
#endif
				rx_lastid=rx_id;					// Aktuelle ID speichern
			}
    	}
	}
	else
	{
#ifdef DEBUG
				uart_putc('n');
#endif
	}
}

void tx_packet(unsigned char retrans)
{
	rf12_stoprx();									// auf TX umschalten
	if ((!retrans)&&((flags&WAITFORACK)||(tx_cnt==0))) // es wird noch auf eine Antwort vom Paket gewartet oder es sind keine Daten zu senden
	{    rf12_txdata(txbuf, 0, tx_status, 0);		// -> kein komplettes neues Paket, sondern nur Status senden
#ifdef DEBUG
		uart_putc('s');
#endif
	}
	else
	{	rf12_txdata(txbuf, tx_cnt, tx_status, tx_id); // komplettes Paket senden
		flags|=WAITFORACK;							// auf ACK warten
		delaycnt=ANSWER_TIMEOUT/2;					// Timeout Counter neu starten
		if (!retrans)								// erstmalige bertragung ?
			retrans_cnt=RETRY;						// -> Retry Counter neu starten
#ifdef DEBUG
		uart_putc('d');
#endif
	}
	rf12_rxmode();									// wieder auf RX umschalten
}

ISR(SIG_OUTPUT_COMPARE1A)							// 500Hz Interrupt
{	
	if (delaycnt)
		delaycnt--;
}



